#pragma once

#include <cmath>
#include <float.h>
#include <stdlib.h>
#include <cassert>

#include "Types.h"

const float kPi = 3.1415926535897932384626433832795f;
const float kDegToRad = kPi/180.0f;
const float kRadToDeg = 180.0f/kPi;

template <typename T>
inline T DegToRad(const T& t)
{
	return t * kDegToRad;
}

template <typename T>
inline T RadToDeg(const T& t)
{
	return t * kRadToDeg;
}

inline float Sin(float theta)
{
	return sinf(theta);
}

inline float Cos(float theta)
{
	return cosf(theta);
}

inline void SinCos(float theta, float& s, float& c)
{
	// no optimizations yet
	s = sinf(theta);
	c = cosf(theta);
}

inline float Tan(float theta)
{
	return tanf(theta);
}

inline float Sqrt(float x)
{
	return sqrtf(x);
}

inline float ASin(float theta)
{
	return asinf(theta);
}

inline float ACos(float theta)
{
	return acosf(theta);
}

inline float ATan(float theta)
{
	return atanf(theta);
}

inline float ATan2(float x, float y)
{
	return atan2f(x, y);
}

inline float Abs(float x)
{
	return fabs(x);
}

inline float Sgn(float x)
{
	return (x < 0.0f ? -1.0f : 1.0f);
}

inline float Sign(float x)
{
	return x < 0.0f ? -1.0f : 1.0f;
}

inline float Min(float a, float b)
{
	return a < b ? a : b;
}

inline float Max(float a, float b)
{
	return a > b ? a : b;
}

inline float Clamp(float a, float low, float high)
{
	return Max(low, Min(a, high));
}

template <typename V, typename T> 
inline T Lerp(const V& start, const V& end, const T& t)
{
	return start + (end-start)*t;
}

template <typename T> 
inline void Swap(T& a, T& b)
{
	T tmp = a;
	a = b;
	b = tmp;
}

inline float RandomUnit()
{
	float r = (float)rand();
	r /= RAND_MAX;
	return r;
}

// Random number in range [-1,1]
inline float RandomSignedUnit()
{
	float r = (float)rand();
	r /= RAND_MAX;
	r = 2.0f * r - 1.0f;
	return r;
}

inline float Random(float lo, float hi)
{
	float r = (float)rand();
	r /= RAND_MAX;
	r = (hi - lo) * r + lo;
	return r;
}

#ifdef _DEBUG
#define VEC2_VALIDATE() {	assert(_finite(x));\
							assert(!_isnan(x));\
											   \
							assert(_finite(y));\
							assert(!_isnan(y));\
						 }
#else
#define VEC2_VALIDATE()
#endif

#ifdef _DEBUG
#define FLOAT_VALIDATE(f) { assert(_finite(f)); assert(!_isnan(f)); }
#else
#define FLOAT_VALIDATE(f)
#endif


// vec2
template <typename T>
class Vector2
{
public:
	
	Vector2() : x(0), y(0) {}
	Vector2(T x_, T y_) : x(x_), y(y_) 
	{
		VEC2_VALIDATE();
	}

	operator T* () const { return &x; }
	operator const T* () const { return &x; };

	void Set(T x_, T y_) { VEC2_VALIDATE(); x = x_; y = y_; }

	Vector2<T> operator * (T scale) const { Vector2<T> r(*this); r *= scale; return r; VEC2_VALIDATE();}
	Vector2<T> operator / (T scale) const { Vector2<T> r(*this); r /= scale; return r; VEC2_VALIDATE();}
	Vector2<T> operator + (const Vector2<T>& v) const { Vector2<T> r(*this); r += v; return r; VEC2_VALIDATE();}
	Vector2<T> operator - (const Vector2<T>& v) const { Vector2<T> r(*this); r -= v; return r; VEC2_VALIDATE();}
	
	Vector2<T>& operator *=(T scale) {x *= scale; y *= scale; VEC2_VALIDATE(); return *this;}
	Vector2<T>& operator /=(T scale) {T s(1.0f/scale); x *= s; y *= s; VEC2_VALIDATE(); return *this;}
	Vector2<T>& operator +=(const Vector2<T>& v) {x += v.x; y += v.y; VEC2_VALIDATE(); return *this;}
	Vector2<T>& operator -=(const Vector2<T>& v) {x -= v.x; y -= v.y; VEC2_VALIDATE(); return *this;}

	// negate
	Vector2<T> operator -() { VEC2_VALIDATE(); return Vector2<T>(-x, -y); }

	// returns this vector
	void Normalize() { *this /= Length(*this); }
	void SafeNormalize(const Vector2<T>& v=Vector2<T>(0.0f,0.0f))
	{
		T length = Length(*this);
		*this = (length==0.0f)?v:(*this /= length);
	}

	T x;
	T y;
};

// lhs scalar scale
template <typename T>
Vector2<T> operator *(T lhs, const Vector2<T>& rhs)
{
	Vector2<T> r(rhs);
	r *= lhs;
	return r;
}

template <typename T>
bool operator==(const Vector2<T>& lhs, const Vector2<T>& rhs)
{
	return (lhs.x == rhs.x && lhs.y == rhs.y);
}


template <typename T>
T Dot(const Vector2<T>& v1, const Vector2<T>& v2)
{
	return v1.x * v2.x + v1.y * v2.y; 
}

template <typename T>
T LengthSq(const Vector2<T>& v)
{
	return Dot(v,v);
}

template <typename T>
T Length(const Vector2<T>& v)
{
	return Sqrt(LengthSq(v));
}

// returns the ccw perpendicular vector 
template <typename T>
Vector2<T> PerpCCW(const Vector2<T>& v)
{
	return Vector2<T>(-v.y, v.x);
}

template <typename T>
Vector2<T> PerpCW(const Vector2<T>& v)
{
	return Vector2<T>(v.y, -v.x);
}

template <typename T>
Vector2<T> Normalize(const Vector2<T>& v)
{
	Vector2<T> a(v);
	a.Normalize();
	return a;
}

template <typename T>
Vector2<T> SafeNormalize(const Vector2<T>& v, const Vector2<T>& fallback=Vector2<T>(0.0f, 0.0f))
{
	Vector2<T> a(v);
	a.SafeNormalize(fallback);
	return a;
}

// component wise min max functions
template <typename T>
Vector2<T> Max(const Vector2<T>& a, const Vector2<T>& b)
{
	return Vector2<T>(std::max(a.x, b.x), std::max(a.y, b.y));
}

template <typename T>
Vector2<T> Min(const Vector2<T>& a, const Vector2<T>& b)
{
	return Vector2<T>(std::min(a.x, b.x), std::min(a.y, b.y));
}

// typedefs
typedef Vector2<float> Vec2;


// a transform represents an 2d affine transformation
template <typename T>
class Matrix23
{
public:

	Matrix23()
	{
		columns[0] = Vector2<T>(1.0f, 0.0f);
		columns[1] = Vector2<T>(0.0f, 1.0f);
		columns[2] = Vector2<T>(0.0f, 0.0f);
	}

	Matrix23(const Vector2<T>& col0, 
			 const Vector2<T>& col1,
			 const Vector2<T>& col2)
	{
		columns[0] = col0;
		columns[1] = col1;
		columns[2] = col2;
	}

	const Vector2<T>& GetCol(uint i) const { return columns[i]; }
	void SetCol(uint i, const Vector2<T>& v) { columns[i] = v; }

	// treats input as a homogenous coordinate with w=1
	Vector2<T> MulPoint(const Vector2<T>& v) const;
	// ignores translation
	Vector2<T> MulVector(const Vector2<T>& v) const;

	Vector2<T> columns[3];
};

template <typename T>
Vector2<T> Matrix23<T>::MulPoint(const Vector2<T>& v) const
{
	return Vector2<T>(columns[0].x*v.x + columns[1].x*v.y + columns[2].x,
					  columns[0].y*v.x + columns[1].y*v.y + columns[2].y);
}

template <typename T>
Vector2<T> Matrix23<T>::MulVector(const Vector2<T>& v) const
{
	return Vector2<T>(columns[0].x*v.x + columns[1].x*v.y,
					  columns[0].y*v.x + columns[1].y*v.y);
}

// global functions
template <typename T>
Vector2<T> MulVector(const Matrix23<T>& m, const Vector2<T>& v)
{
	return m.MulVector(v);
}

template <typename T>
Vector2<T> MulPoint(const Matrix23<T>& m, const Vector2<T>& v)
{
	return m.MulPoint(v);
}


typedef Matrix23<float> Mat23;

// creates a vector given an angle measured clockwise from vertical (0,1)
inline Vec2 AngleToVector(float a)
{
	return Vec2(Cos(a), Sin(a));
}

// hermite spline interpolation
template <typename T>
T HermiteInterpolate(const T& a, const T& b, const T& t1, const T& t2, float t)
{
	// blending weights
	const float w1 = 1.0f - 3*t*t + 2*t*t*t;
	const float w2 = t*t*(3.0f-2.0f*t);
	const float w3 = t*t*t - 2*t*t + t;
	const float w4 = t*t*(t-1.0f);

	// return weighted combination
	return a*w1 + b*w2 + t1*w3 + t2*w4;

}
	
template <typename T>
T HermiteTangent(const T& a, const T& b, const T& t1, const T& t2, float t)
{
	// first derivative blend weights
	const float w1 = 6.0f*t*(t-1.0f);
	const float w2 = 6.0f*t*(1.0f-t);
	const float w3 = 1.0f-4.0f*t+3.0f*t*t;
	const float w4 = 2.0f*t + 3.0f*t*t;

	// weighted combination
	return a*w1 + b*w2 + t1*w3 + t2*w4;
}

// simple colour class
class Colour
{
public:

	enum Preset
	{
		kRed,
		kGreen,
		kBlue,
		kWhite,
		kBlack
	};
	
	Colour(float r_=0.0f, float g_=0.0f, float b_=0.0f, float a_=1.0f) : r(r_), g(g_), b(b_), a(a_) {}
	Colour(Preset p);

	// cast operator
	operator const float*() const { return &r; }
	operator float*() { return &r; }

	float r, g, b, a;
};